"""
Managers for AA Forum models
"""

# pylint: disable=cyclic-import

# Django
from django.contrib.auth.models import User
from django.db import models
from django.db.models import Prefetch, Q, QuerySet


class SettingQuerySet(models.QuerySet):
    """
    SettingQuerySet
    """

    def delete(self):
        """
        Delete action

        Override :: We don't allow deletion here, so we make sure the object
                    is saved again and not deleted

        :return:
        :rtype:
        """

        return super().update()


class SettingManager(models.Manager):
    """
    SettingManager
    """

    def get_setting(self, setting_key: str) -> str:
        """
        Get a setting value

        :param setting_key:
        :type setting_key:
        :return:
        :rtype:
        """

        return getattr(self.first(), setting_key)

    def get_queryset(self):
        """
        Get a Setting queryset

        :return:
        :rtype:
        """

        return SettingQuerySet(self.model)


class BoardQuerySet(models.QuerySet):
    """
    BoardQuerySet
    """

    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter boards that given user has access to.

        :param user:
        :type user:
        :return:
        :rtype:
        """

        # Forum manager always has access, so assign this permission wisely
        if user.has_perm(perm="aa_forum.manage_forum"):
            return self

        # If not a forum manager, check if the user has access to the board
        return self.filter(
            Q(groups__in=user.groups.all()) | Q(groups__isnull=True)
        ).distinct()


class BoardManagerBase(models.Manager):
    """
    BoardManagerBase
    """

    pass  # pylint: disable=unnecessary-pass


BoardManager = BoardManagerBase.from_queryset(BoardQuerySet)


class TopicQuerySet(models.QuerySet):
    """
    TopicQuerySet
    """

    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter boards that given user has access to.

        :param user:
        :type user:
        :return:
        :rtype:
        """

        # Forum manager always has access, so assign this permission wisely
        if user.has_perm(perm="aa_forum.manage_forum"):
            return self

        # If not a forum manager, check if the user has access to the board
        return self.filter(
            Q(board__groups__in=user.groups.all()) | Q(board__groups__isnull=True)
        ).distinct()

    def get_from_slugs(
        self,
        category_slug: str,
        board_slug: str,
        topic_slug: str,
        user: User,
    ) -> models.Model:
        """
        Fetch a topic from slugs for user. Return None if not found or no access.

        :param category_slug:
        :type category_slug:
        :param board_slug:
        :type board_slug:
        :param topic_slug:
        :type topic_slug:
        :param user:
        :type user:
        :return:
        :rtype:
        """

        # AA Forum
        from aa_forum.models import Message  # pylint: disable=import-outside-toplevel

        try:
            topic = (
                self.select_related(
                    "board",
                    "board__category",
                    "first_message",
                    "first_message__topic",
                    "first_message__topic__board",
                    "first_message__topic__board__category",
                )
                .prefetch_related(
                    Prefetch(
                        lookup="messages",
                        queryset=Message.objects.select_related(
                            "user_created",
                            "user_created__profile__main_character",
                            "user_created__aa_forum_user_profile",
                        ).order_by("time_posted"),
                        to_attr="messages_sorted",
                    )
                )
                .filter(
                    board__category__slug=str(category_slug),
                    board__slug=str(board_slug),
                    slug=str(topic_slug),
                )
                .user_has_access(user=user)
                .distinct()
                .get()
            )
        except self.model.DoesNotExist:
            return None

        return topic


class TopicManagerBase(models.Manager):
    """
    TopicManagerBase
    """

    pass  # pylint: disable=unnecessary-pass


TopicManager = TopicManagerBase.from_queryset(TopicQuerySet)


class MessageQuerySet(models.QuerySet):
    """
    MessageQuerySet
    """

    def user_has_access(self, user: User) -> models.QuerySet:
        """
        Filter the topics a given user has access to.

        :param user:
        :type user:
        :return:
        :rtype:
        """

        # Forum manager always has access, so assign this permission wisely
        if user.has_perm(perm="aa_forum.manage_forum"):
            return self

        # If not a forum manager, check if the user has access to the board
        return self.filter(
            Q(topic__board__groups__in=user.groups.all())
            | Q(topic__board__groups__isnull=True)
        ).distinct()

    def get_from_slugs(  # pylint: disable=too-many-arguments
        self,
        category_slug: str,
        board_slug: str,
        topic_slug: str,
        message_id: int,
        user: User,
    ) -> models.Model:
        """
        Fetch a message from slugs for user. Return None if not found or no access.

        :param category_slug:
        :type category_slug:
        :param board_slug:
        :type board_slug:
        :param topic_slug:
        :type topic_slug:
        :param message_id:
        :type message_id:
        :param user:
        :type user:
        :return:
        :rtype:
        """

        try:
            message = (
                self.select_related(
                    "topic",
                    "topic__board",
                    "topic__board__category",
                    "topic__first_message",
                    "topic__first_message__topic",
                    "topic__first_message__topic__board",
                    "topic__first_message__topic__board__category",
                )
                .filter(
                    topic__board__category__slug=str(category_slug),
                    topic__board__slug=str(board_slug),
                    topic__slug=str(topic_slug),
                    pk=message_id,
                )
                .user_has_access(user=user)
                .distinct()
                .get()
            )
        except self.model.DoesNotExist:
            return None

        return message


class MessageManagerBase(models.Manager):
    """
    MessageManagerBase
    """

    pass  # pylint: disable=unnecessary-pass


MessageManager = MessageManagerBase.from_queryset(MessageQuerySet)


class PersonalMessageQuerySet(models.QuerySet):
    """
    PersonalMessageQuerySet
    """

    def get_personal_messages_for_user(self, user: User) -> QuerySet:
        """
        Get a user's personal messages

        :param user:
        :type user:
        :return:
        :rtype:
        """

        messages = (
            self.filter(recipient=user, deleted_by_recipient=False)
            .select_related(
                "sender", "sender__profile", "sender__profile__main_character"
            )
            .order_by("-time_sent")
        )

        return messages

    def get_personal_messages_sent_for_user(self, user: User) -> QuerySet:
        """
        Get a user's personal messages sent

        :param user:
        :type user:
        :return:
        :rtype:
        """

        messages = (
            self.filter(sender=user, deleted_by_sender=False)
            .select_related(
                "recipient", "recipient__profile", "recipient__profile__main_character"
            )
            .order_by("-time_sent")
        )

        return messages

    def get_personal_message_unread_count_for_user(self, user: User) -> int:
        """
        Get a user's unread personal messages count

        :param user:
        :type user:
        :return:
        :rtype:
        """

        unread_count = self.filter(
            recipient=user, is_read=False, deleted_by_recipient=False
        ).count()

        return unread_count


class PersonalMessageManagerBase(models.Manager):
    """
    PersonalMessageManagerBase
    """

    pass  # pylint: disable=unnecessary-pass


PersonalMessageManager = PersonalMessageManagerBase.from_queryset(
    PersonalMessageQuerySet
)



# Generated by Django 3.2.4 on 2021-06-28 15:55

# Django
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

# AA Forum
import aa_forum.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "verbose_name": "AA-Forum",
                "permissions": (
                    ("basic_access", "Can access the AA-Forum module"),
                    (
                        "manage_forum",
                        "Can manage the AA-Forum module (Category, topics and messages)",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Board",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=254)),
                (
                    "slug",
                    models.SlugField(allow_unicode=True, max_length=254, unique=True),
                ),
                ("description", models.TextField(blank=True)),
                ("order", models.IntegerField(db_index=True, default=999999)),
            ],
            options={
                "verbose_name": "board",
                "verbose_name_plural": "boards",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Category",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=254, unique=True)),
                (
                    "slug",
                    models.SlugField(allow_unicode=True, max_length=254, unique=True),
                ),
                ("is_collapsible", models.BooleanField(default=False)),
                ("order", models.IntegerField(db_index=True, default=999999)),
            ],
            options={
                "verbose_name": "category",
                "verbose_name_plural": "categories",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Message",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("time_posted", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("time_modified", models.DateTimeField(auto_now=True)),
                ("message", models.TextField(blank=True)),
                ("message_plaintext", models.TextField(blank=True)),
            ],
            options={
                "verbose_name": "message",
                "verbose_name_plural": "messages",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("variable", models.CharField(max_length=254, unique=True)),
                ("value", models.TextField()),
            ],
            options={
                "verbose_name": "setting",
                "verbose_name_plural": "settings",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Topic",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("subject", models.CharField(max_length=254)),
                (
                    "slug",
                    models.SlugField(allow_unicode=True, max_length=254, unique=True),
                ),
                ("is_sticky", models.BooleanField(db_index=True, default=False)),
                ("is_locked", models.BooleanField(db_index=True, default=False)),
                (
                    "board",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="topics",
                        to="aa_forum.board",
                    ),
                ),
                (
                    "first_message",
                    models.ForeignKey(
                        default=None,
                        editable=False,
                        help_text="Shortcut for better performance",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="aa_forum.message",
                    ),
                ),
                (
                    "last_message",
                    models.ForeignKey(
                        default=None,
                        editable=False,
                        help_text="Shortcut for better performance",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="aa_forum.message",
                    ),
                ),
            ],
            options={
                "verbose_name": "topic",
                "verbose_name_plural": "topics",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="PersonalMessage",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("time_sent", models.DateTimeField(auto_now_add=True)),
                ("subject", models.CharField(max_length=254)),
                (
                    "slug",
                    models.SlugField(allow_unicode=True, max_length=254, unique=True),
                ),
                ("message", models.TextField(blank=True)),
                ("is_read", models.BooleanField(default=False)),
                (
                    "recipient",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "sender",
                    models.ForeignKey(
                        on_delete=models.SET(aa_forum.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "personal message",
                "verbose_name_plural": "personal messages",
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="message",
            name="topic",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="messages",
                to="aa_forum.topic",
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="user_created",
            field=models.ForeignKey(
                on_delete=models.SET(aa_forum.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="message",
            name="user_updated",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=models.SET(aa_forum.models.get_sentinel_user),
                related_name="+",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="LastMessageSeen",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("message_time", models.DateTimeField()),
                (
                    "topic",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="aa_forum.topic"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.AddField(
            model_name="board",
            name="category",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="boards",
                to="aa_forum.category",
            ),
        ),
        migrations.AddField(
            model_name="board",
            name="first_message",
            field=models.ForeignKey(
                default=None,
                editable=False,
                help_text="Shortcut for better performance",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="aa_forum.message",
            ),
        ),
        migrations.AddField(
            model_name="board",
            name="groups",
            field=models.ManyToManyField(
                blank=True,
                related_name="aa_forum_boards_group_restriction",
                to="auth.Group",
            ),
        ),
        migrations.AddField(
            model_name="board",
            name="last_message",
            field=models.ForeignKey(
                default=None,
                editable=False,
                help_text="Shortcut for better performance",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="aa_forum.message",
            ),
        ),
        migrations.AddField(
            model_name="board",
            name="parent_board",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="child_boards",
                to="aa_forum.board",
            ),
        ),
        migrations.AddConstraint(
            model_name="topic",
            constraint=models.UniqueConstraint(
                fields=("board", "subject"), name="fpk_topic"
            ),
        ),
        migrations.AddIndex(
            model_name="lastmessageseen",
            index=models.Index(
                fields=["topic", "user", "message_time"],
                name="lastmessageseen_compounded",
            ),
        ),
        migrations.AddConstraint(
            model_name="board",
            constraint=models.UniqueConstraint(
                fields=("category", "name"), name="fpk_board"
            ),
        ),
    ]



"""
AA Forum template tags
"""

# Standard Library
import re
from datetime import datetime

# Third Party
from bs4 import BeautifulSoup

# Django
from django import template
from django.contrib.auth.models import User
from django.template.defaulttags import register
from django.templatetags.static import static
from django.utils import formats
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth (External Libs)
from app_utils.urls import reverse as reverse_url

# AA Forum
from aa_forum import __version__
from aa_forum.app_settings import aa_timezones_installed
from aa_forum.constants import SEARCH_STOPWORDS
from aa_forum.models import PersonalMessage


class SetVarNode(template.Node):
    """
    Set a template variable
    """

    def __init__(self, var_name, var_value):
        self.var_name = var_name
        self.var_value = var_value

    def render(self, context):
        """
        Render the template variable

        :param context:
        :type context:
        :return:
        :rtype:
        """

        try:
            value = template.Variable(var=self.var_value).resolve(context=context)
        except template.VariableDoesNotExist:
            value = ""

        context[self.var_name] = value

        return ""


@register.simple_tag
def aa_forum_static(path: str) -> str:
    """
    Return a versioned static URL

    :param path:
    :type path:
    :return:
    :rtype:
    """

    static_url = static(path=path)
    versioned_url = static_url + "?v=" + __version__

    return versioned_url


@register.filter
def aa_forum_time(db_datetime: datetime) -> str:
    """
    Format a datetime object for the forum

    :param db_datetime:
    :type db_datetime:
    :return:
    :rtype:
    """

    # If empty, return empty
    if db_datetime in (None, ""):
        return ""

    # Try to format the date for a localized output
    try:
        formatted_date_string = formats.date_format(value=db_datetime)
    except AttributeError:
        try:
            formatted_date_string = format(db_datetime)
        except AttributeError:
            formatted_date_string = ""

    formatted_time_string = db_datetime.strftime("%H:%M:%S")
    formatted_forum_date = f"{formatted_date_string}, {formatted_time_string}"

    # If `aa-timezones` is installed, add (?) to the date-time string
    # and link to the time zones conversion
    if aa_timezones_installed():
        timestamp_from_db_datetime = int(datetime.timestamp(db_datetime))
        timezones_url = reverse_url(
            viewname="timezones:index", args=[timestamp_from_db_datetime]
        )
        link_title = _("Timezone conversion")

        return mark_safe(
            s=(
                f"{formatted_forum_date} "
                f'<sup>(<a href="{timezones_url}" target="_blank" rel="noopener noreferer" '
                f'title="{link_title}" data-bs-tooltip="aa-forum">'
                '<i class="fa-solid fa-circle-question"></i></a>)</sup>'
            )
        )

    return formatted_forum_date


@register.filter
def aa_forum_highlight_search_term(text: str, search_phrase: str) -> str:
    """
    Highlight a search term in a text

    :param text:
    :type text:
    :param search_phrase:
    :type search_phrase:
    :return:
    :rtype:
    """

    delimiter_search_term_start = "{«}"
    delimiter_search_term_end = "{»}"

    querywords = search_phrase.split()
    search_phrase_terms = [
        word for word in querywords if word.lower() not in SEARCH_STOPWORDS
    ]
    highlighted = text

    for search_term in search_phrase_terms:
        highlighted = re.sub(
            pattern=f"(?i)({(re.escape(search_term))})",
            repl=f"{delimiter_search_term_start}\\1{delimiter_search_term_end}",
            string=highlighted,
        )

    highlighted = BeautifulSoup(markup=highlighted, features="html.parser")

    for hyperlink in highlighted.findAll(name="a"):
        try:
            hyperlink["href"] = (
                hyperlink["href"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            # this should never happen, but in case it does, add a dummy href
            hyperlink["href"] = "#"

        try:
            hyperlink["title"] = (
                hyperlink["title"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            pass

        try:
            hyperlink["name"] = (
                hyperlink["name"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            pass

    for image in highlighted.findAll("img"):
        try:
            image["src"] = (
                image["src"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            # this should never happen, but in case it does, add a dummy src
            image["src"] = "#"

        try:
            image["alt"] = (
                image["alt"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            pass

        try:
            image["title"] = (
                image["title"]
                .replace(delimiter_search_term_start, "")
                .replace(delimiter_search_term_end, "")
            )
        except KeyError:
            pass

    return mark_safe(
        str(highlighted)
        .replace(
            delimiter_search_term_start, '<span class="aa-forum-search-term-highlight">'
        )
        .replace(delimiter_search_term_end, "</span>")
    )


@register.tag(name="aa_forum_template_variable")
def aa_forum_template_variable(parser, token):  # pylint: disable=unused-argument
    """
    Set a template variable

    Usage:
        {% aa_forum_template_variable <var_name> = <var_value> %}

    :param parser:
    :type parser:
    :param token:
    :type token:
    :return:
    :rtype:
    """

    parts = token.split_contents()

    if len(parts) < 4:
        raise template.TemplateSyntaxError(
            "'aa_forum_template_variable' tag must be of the form: "
            "{% aa_forum_template_variable <var_name> = <var_value> %}"
        )

    return SetVarNode(parts[1], parts[3])


@register.simple_tag
def personal_message_unread_count(user: User) -> str:
    """
    Get the unread personal message count for a user

    :param user:
    :type user:
    :return:
    :rtype:
    """

    return_value = ""
    message_count = PersonalMessage.objects.get_personal_message_unread_count_for_user(
        user=user
    )

    if message_count > 0:
        return_value = mark_safe(
            s=f'<span class="badge bg-light aa-forum-badge-personal-messages-unread-count">{message_count}</span>'  # pylint: disable=line-too-long
        )

    return return_value


@register.filter
def aa_forum_main_character_name(user: User) -> str:
    """
    Get the users main character name, or return the username if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.character_name
    except AttributeError:
        return str(user)

    return return_value


# pylint: disable=duplicate-code
@register.filter
def aa_forum_main_character_id(user: User) -> int:
    """
    Get the users' main character id, or 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.character_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def aa_forum_main_character_corporation_name(user: User) -> str:
    """
    Get the users' main character corporation name, or an empty string if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.corporation_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def aa_forum_main_character_corporation_id(user: User) -> int:
    """
    Get the users' main character corporation id, or 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.corporation_id
    except AttributeError:
        return_value = 1

    return return_value


@register.filter
def aa_forum_main_character_alliance_name(user: User) -> str:
    """
    Get the users' main character alliance name, or an empty string if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return ""

    try:
        return_value = user.profile.main_character.alliance_name
    except AttributeError:
        return_value = ""

    return return_value


@register.filter
def aa_forum_main_character_alliance_id(user: User) -> int:
    """
    Get the users' main character alliance id, or 1 if no main character

    :param user:
    :type user:
    :return:
    :rtype:
    """

    if user is None:
        return 1

    try:
        return_value = user.profile.main_character.alliance_id

        # Check if the user is in an alliance
        try:
            return_value = int(return_value)
        except Exception:  # pylint: disable=broad-exception-caught
            return_value = 1
    except AttributeError:
        return_value = 1

    return return_value



"""
Forum views
"""

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.auth.models import Group
from django.core.handlers.wsgi import WSGIRequest
from django.db.models import Count, Exists, OuterRef, Prefetch, Q
from django.http import HttpResponse, HttpResponseNotFound, HttpResponseRedirect
from django.shortcuts import redirect, render
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA Forum
from aa_forum import __title__
from aa_forum.forms import EditMessageForm, EditTopicForm, NewTopicForm
from aa_forum.helper.discord_messages import send_message_to_discord_webhook
from aa_forum.helper.pagination import get_paginated_page_object
from aa_forum.models import Board, Category, LastMessageSeen, Message, Setting, Topic

logger = LoggerAddTag(my_logger=get_extension_logger(name=__name__), prefix=__title__)


@login_required
@permission_required(perm="aa_forum.basic_access")
def index(request: WSGIRequest) -> HttpResponse:
    """
    Forum index view

    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(
        ~Exists(queryset=has_read_all_messages)
    ).values_list("pk", flat=True)

    boards = (
        Board.objects.select_related(
            "category",
            "last_message",
            "last_message__topic",
            "last_message__user_created__profile__main_character",
            "first_message",
        )
        .prefetch_related(
            Prefetch(
                lookup="child_boards",
                queryset=Board.objects.select_related(
                    "category",
                    "parent_board",
                    "last_message",
                    "last_message__user_created",
                    "last_message__user_created__profile__main_character",
                    "first_message",
                    "first_message__user_created",
                    "first_message__user_created__profile__main_character",
                )
                .annotate(
                    num_posts=Count(expression="topics__messages", distinct=True),
                    num_topics=Count(expression="topics", distinct=True),
                    num_unread=Count(
                        expression="topics",
                        filter=Q(topics__in=unread_topic_pks),
                        distinct=True,
                    ),
                )
                .user_has_access(user=request.user)
                .order_by("order", "id"),
            )
        )
        .prefetch_related(
            Prefetch(lookup="groups", queryset=Group.objects.order_by("name"))
        )
        .prefetch_related("topics")
        .user_has_access(user=request.user)
        .filter(parent_board__isnull=True)
        .annotate(
            num_posts=Count(expression="topics__messages", distinct=True),
            num_topics=Count(expression="topics", distinct=True),
            num_unread=Count(
                expression="topics",
                filter=Q(topics__in=unread_topic_pks),
                distinct=True,
            ),
        )
        .order_by("category__order", "category__id", "order", "id")
    )

    categories_map = {}

    for board_in_loop in boards:
        category = board_in_loop.category

        if category.pk not in categories_map:
            categories_map[category.pk] = {
                "id": category.id,
                "name": category.name,
                "boards_sorted": [],
                "order": category.order,
            }

        categories_map[category.pk]["boards_sorted"].append(board_in_loop)

    categories = sorted(categories_map.values(), key=lambda k: k["order"])
    context = {"categories": categories}

    logger.info(msg=f"{request.user} called forum index.")

    return render(
        request=request, template_name="aa_forum/view/forum/index.html", context=context
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def board(
    request: WSGIRequest, category_slug: str, board_slug: str, page_number: int = None
) -> HttpResponse:
    """
    View a board

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(
        ~Exists(queryset=has_read_all_messages)
    ).values_list("pk", flat=True)

    try:
        current_board = (
            Board.objects.select_related("category")
            .select_related("parent_board")
            .prefetch_related(
                Prefetch(
                    lookup="child_boards",
                    queryset=Board.objects.select_related(
                        "category",
                        "parent_board",
                        "last_message",
                        "last_message__user_created",
                        "last_message__user_created__profile__main_character",
                        "first_message",
                        "first_message__user_created",
                        "first_message__user_created__profile__main_character",
                    )
                    .prefetch_related(
                        Prefetch(
                            lookup="groups", queryset=Group.objects.order_by("name")
                        )
                    )
                    .prefetch_related("topics")
                    .annotate(
                        num_posts=Count(expression="topics__messages", distinct=True),
                        num_topics=Count(expression="topics", distinct=True),
                        num_unread=Count(
                            expression="topics",
                            filter=Q(topics__in=unread_topic_pks),
                            distinct=True,
                        ),
                    )
                    .user_has_access(user=request.user)
                    .order_by("order", "id"),
                )
            )
            .prefetch_related(
                Prefetch(
                    lookup="topics",
                    queryset=Topic.objects.select_related(
                        "last_message",
                        "last_message__user_created",
                        "last_message__user_created__profile__main_character",
                        "first_message",
                        "first_message__user_created",
                        "first_message__user_created__profile__main_character",
                    )
                    .annotate(num_posts=Count(expression="messages", distinct=True))
                    .annotate(
                        has_unread_messages=~Exists(queryset=has_read_all_messages)
                    )
                    .order_by("-is_sticky", "-last_message__time_posted", "-id"),
                    to_attr="topics_sorted",
                )
            )
            .prefetch_related(
                Prefetch(
                    lookup="announcement_groups",
                    queryset=Group.objects.order_by("name"),
                )
            )
            .filter(category__slug=category_slug, slug=board_slug)
            .user_has_access(user=request.user)
            .get()
        )
    except Board.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The board you were trying to visit does "
                    "either not exist, or you don't have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} called board without having access to it. "
                "Redirecting to forum index."
            )
        )

        return redirect(to="aa_forum:forum_index")

    page_obj = get_paginated_page_object(
        queryset=current_board.topics_sorted,
        items_per_page=Setting.objects.get_setting(
            setting_key=Setting.Field.TOPICSPERPAGE
        ),
        page_number=page_number,
    )

    context = {
        "board": current_board,
        "user_can_start_topic": current_board.user_can_start_topic(user=request.user),
        "page_obj": page_obj,
    }

    logger.info(msg=f'{request.user} called board "{current_board.name}".')

    return render(
        request=request, template_name="aa_forum/view/forum/board.html", context=context
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def board_new_topic(
    request: WSGIRequest, category_slug: str, board_slug: str
) -> HttpResponse:
    """
    Create a new topic in a board

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :return:
    :rtype:
    """

    try:
        Category.objects.get(slug__exact=category_slug)
    except Category.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The category you were trying to post in "
                    "does not exist.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} tried to open a non existing category. "
                "Redirecting to forum index."
            )
        )

        return redirect(to="aa_forum:forum_index")

    try:
        current_board: Board = (
            Board.objects.select_related("category")
            .user_has_access(user=request.user)
            .filter(category__slug=category_slug, slug=board_slug)
            .get()
        )
    except Board.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The board you were trying to post in does "
                    "either not exist, or you don't have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} tried to create a topic in a board they have no "
                "access to. Redirecting to forum index."
            )
        )

        return redirect(to="aa_forum:forum_index")

    if not current_board.user_can_start_topic(user=request.user):
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The board you were trying to post in is "
                    "an announcement board and you don't have the permissions to "
                    "start a topic there.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} tried to create a topic in an announcement board "
                "without the permission to do so. Redirecting to board index."
            )
        )

        return redirect(
            to="aa_forum:forum_board",
            category_slug=current_board.category.slug,
            board_slug=current_board.slug,
        )

    # If this is a POST request, we need to process the form data …
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = NewTopicForm(data=request.POST)

        # Check whether it's valid:
        if form.is_valid():
            try:
                # Let's see if we can create a new topic
                new_topic = current_board.new_topic(
                    subject=form.cleaned_data["subject"],
                    message=form.cleaned_data["message"],
                    user=request.user,
                )
            except current_board.TopicAlreadyExists as exc:
                # Apparently there is already a topic with this subject
                logger.debug(msg=f"{request.user} tried to create a duplicate topic.")

                messages.warning(request=request, message=exc)

                return render(
                    request=request,
                    template_name="aa_forum/view/forum/new-topic.html",
                    context={"board": current_board, "form": form},
                )

            return redirect(
                to="aa_forum:forum_topic",
                category_slug=current_board.category.slug,
                board_slug=current_board.slug,
                topic_slug=new_topic.slug,
            )

        # The form is invalid
        messages.error(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    "<h4>Error!</h4>"
                    "<p>Either subject or message is missing. "
                    "Please make sure you enter both fields, "
                    "as both fields are mandatory.</p>"
                )
            ),
        )
    # If not, we'll create a blank form
    else:
        form = NewTopicForm()

    context = {"board": current_board, "form": form}

    logger.info(
        msg=f'{request.user} is starting a new topic in board "{current_board.name}".'
    )

    return render(
        request=request,
        template_name="aa_forum/view/forum/new-topic.html",
        context=context,
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def topic(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    page_number: int = None,
) -> HttpResponse:
    """
    View a topic

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    current_topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not current_topic:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The topic you were trying to view does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} called a non existent topic. Redirecting to forum index."  # pylint: disable=line-too-long
            )
        )

        return redirect(to="aa_forum:forum_index")

    # Determine if the current user can modify the topics subject
    can_modify_subject = False
    if (
        request.user == current_topic.first_message.user_created
        or request.user.has_perm(perm="aa_forum.manage_forum")
    ):
        can_modify_subject = True

    page_obj = get_paginated_page_object(
        queryset=current_topic.messages_sorted,
        items_per_page=Setting.objects.get_setting(
            setting_key=Setting.Field.MESSAGESPERPAGE
        ),
        page_number=page_number,
    )

    # Set this topic as "read by" by the current user
    try:
        last_message_on_page = page_obj.object_list[-1]
    except IndexError:
        pass
    else:
        try:
            last_message_seen = LastMessageSeen.objects.get(
                topic=current_topic, user=request.user
            )
        except LastMessageSeen.DoesNotExist:
            last_message_seen = None

        if (
            not last_message_seen
            or last_message_seen.message_time < last_message_on_page.time_posted
        ):
            LastMessageSeen.objects.update_or_create(
                topic=current_topic,
                user=request.user,
                defaults={"message_time": last_message_on_page.time_posted},
            )

    context = {
        "topic": current_topic,
        "can_modify_subject": can_modify_subject,
        "page_obj": page_obj,
        "reply_form": EditMessageForm(),
    }

    logger.info(msg=f'{request.user} called topic "{current_topic.subject}".')

    return render(
        request=request, template_name="aa_forum/view/forum/topic.html", context=context
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def topic_modify(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
) -> HttpResponse:
    """
    Modify a topic subject

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    topic_to_modify = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not topic_to_modify:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The topic you were trying to modify does "
                    "not exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} called a non existent topic. Redirecting to forum index."  # pylint: disable=line-too-long
            )
        )

        return redirect(to="aa_forum:forum_index")

    # Check if the user actually has the right to edit this message
    if (
        topic_to_modify.first_message.user_created != request.user
        and not request.user.has_perm(perm="aa_forum.manage_forum")
    ):
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>You are not allowed to modify this topic!</p>")
            ),
        )

        logger.info(
            msg=(
                f'{request.user} tried to modify topic "{topic_to_modify.subject}" '
                "without the proper permissions. Redirecting to forum index."
            )
        )

        return redirect(
            to="aa_forum:forum_topic",
            category_slug=category_slug,
            board_slug=board_slug,
            topic_slug=topic_slug,
        )

    # We are in the clear, let's see what we've got
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditTopicForm(data=request.POST)

        # Check whether it's valid:
        if form.is_valid():
            topic_to_modify.subject = form.cleaned_data["subject"]
            topic_to_modify.save()

            messages.success(
                request=request,
                message=mark_safe(
                    # pylint: disable=duplicate-code
                    s=_("<h4>Success!</h4><p>The topic subject has been updated.</p>")
                ),
            )

            logger.info(
                msg=f'{request.user} modified topic "{topic_to_modify.subject}".'
            )

            return redirect(
                to="aa_forum:forum_topic",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
            )
    # If not, we'll fill the form with the information from the message object
    else:
        form = EditTopicForm(instance=topic_to_modify)

    context = {"form": form, "topic": topic_to_modify}

    logger.info(msg=f'{request.user} modifying "{topic_to_modify.subject}".')

    return render(
        request=request,
        template_name="aa_forum/view/forum/modify-topic.html",
        context=context,
    )


def _topic_from_slugs(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> Topic | None:
    """
    Helper function to get a topic from slugs

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    current_topic = Topic.objects.get_from_slugs(
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        user=request.user,
    )

    return current_topic


@login_required
@permission_required(perm="aa_forum.basic_access")
def topic_first_unread_message(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> HttpResponse:
    """
    Redirect to the first unread message in a topic

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    current_topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not current_topic:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The topic you were trying to view does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        return redirect(to="aa_forum:forum_index")

    messages_sorted = current_topic.messages.order_by("time_posted")

    try:
        last_message_seen = LastMessageSeen.objects.filter(
            topic=current_topic, user=request.user
        ).get()
    except LastMessageSeen.DoesNotExist:
        redirect_message = messages_sorted.first()
    else:
        redirect_message = messages_sorted.filter(
            time_posted__gt=last_message_seen.message_time
        ).first()

        if not redirect_message:
            redirect_message = messages_sorted.last()

    if redirect_message:
        return redirect(to=redirect_message.get_absolute_url())

    return redirect(to=current_topic.get_absolute_url())


def _get_boards_with_unread_topics(request: WSGIRequest):
    """
    Get all boards with unread topics

    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(
        ~Exists(queryset=has_read_all_messages)
    ).values_list("pk", flat=True)

    boards = (
        Board.objects.select_related(
            "category",
            "last_message",
            "last_message__topic",
            "last_message__user_created__profile__main_character",
            "first_message",
        )
        .prefetch_related(
            Prefetch(
                lookup="topics",
                queryset=Topic.objects.select_related(
                    "last_message",
                    "last_message__user_created",
                    "last_message__user_created__profile__main_character",
                    "first_message",
                    "first_message__user_created",
                    "first_message__user_created__profile__main_character",
                )
                .filter(pk__in=unread_topic_pks)
                .annotate(num_posts=Count(expression="messages", distinct=True))
                .annotate(has_unread_messages=~Exists(queryset=has_read_all_messages))
                .order_by("-is_sticky", "-last_message__time_posted", "-id"),
            )
        )
        .filter(topics__in=unread_topic_pks)
        .user_has_access(user=request.user)
        .order_by("category__order", "category__id", "order", "id")
        .all()
        .distinct()
    )

    return boards


@login_required
@permission_required(perm="aa_forum.basic_access")
def topic_show_all_unread(request: WSGIRequest) -> HttpResponse:
    """
    Show all unread topics

    :param request:
    :type request:
    :return:
    :rtype:
    """

    context = {"boards": _get_boards_with_unread_topics(request=request)}

    logger.info(msg=f"{request.user} calling unread topics view.")

    return render(
        request=request,
        template_name="aa_forum/view/forum/unread-topics.html",
        context=context,
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def topic_reply(
    request: WSGIRequest, category_slug: str, board_slug: str, topic_slug: str
) -> HttpResponse:
    """
    Reply to a topic

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :return:
    :rtype:
    """

    current_topic = _topic_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )

    if not current_topic:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The topic you were trying to reply does not "
                    "exist or you do not have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} called a non existent topic. Redirecting to forum index."  # pylint: disable=line-too-long
            )
        )

        return redirect(to="aa_forum:forum_index")

    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(data=request.POST)

        # Check whether it's valid:
        if form.is_valid():
            new_message = Message(
                topic=current_topic,
                user_created=request.user,
                message=form.cleaned_data["message"],
            )
            new_message.save()

            # Check if the user has the rights to close or reopen the topic
            if (
                request.user.has_perm(perm="aa_forum.manage_forum")
                or request.user == current_topic.first_message.user_created
            ):
                # Close the topic if requested
                if (
                    form.cleaned_data["close_topic"]
                    and not form.cleaned_data["reopen_topic"]
                ):
                    current_topic.is_locked = True
                    current_topic.save(update_fields=["is_locked"])

                # Reopen the topic if requested
                if (
                    form.cleaned_data["reopen_topic"]
                    and not form.cleaned_data["close_topic"]
                ) and request.user.has_perm(perm="aa_forum.manage_forum"):
                    current_topic.is_locked = False
                    current_topic.save(update_fields=["is_locked"])

            # Send to webhook if one is configured
            if (
                current_topic.board.discord_webhook is not None
                and current_topic.board.use_webhook_for_replies is not False
            ):
                send_message_to_discord_webhook(
                    board=current_topic.board,
                    topic=current_topic,
                    message=new_message,
                    headline=f'**New reply has been posted in topic "{current_topic.subject}"**',
                )

            logger.info(
                msg=f'{request.user} replied to topic "{current_topic.subject}".'
            )

            return redirect(
                to="aa_forum:forum_message",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
                message_id=new_message.id,
            )

        messages.error(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    "<h4>Error!</h4>"
                    "<p>Message field is mandatory and cannot be empty.</p>"
                )
            ),
        )
    else:
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>Something went wrong, please try again.</p>")
            ),
        )

    return redirect(
        to="aa_forum:forum_topic",
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
    )


@login_required
@permission_required(perm="aa_forum.manage_forum")
def topic_change_lock_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the lock state of the given topic

    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        current_topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    if current_topic.is_locked:
        current_topic.is_locked = False

        messages.success(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    f'<h4>Success!</h4><p>Topic "{current_topic}" has been unlocked/re-opened.</p>'  # pylint: disable=line-too-long
                )
            ),
        )

        logger.info(msg=f'{request.user} unlocked/re-opened topic "{current_topic}".')
    else:
        current_topic.is_locked = True

        messages.success(
            request=request,
            message=mark_safe(
                # pylint: disable=duplicate-code
                s=_(
                    f'<h4>Success!</h4><p>Topic "{current_topic}" has been locked/closed.</p>'  # pylint: disable=line-too-long
                )
            ),
        )

        logger.info(msg=f'{request.user} locked/closed "{current_topic}".')

    current_topic.save(update_fields=["is_locked"])

    return redirect(
        to="aa_forum:forum_board",
        category_slug=current_topic.board.category.slug,
        board_slug=current_topic.board.slug,
    )


@login_required
@permission_required(perm="aa_forum.manage_forum")
def topic_change_sticky_state(
    request: WSGIRequest, topic_id: int
) -> HttpResponseRedirect:
    """
    Change the sticky state of the given topic

    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        curent_topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    if curent_topic.is_sticky:
        curent_topic.is_sticky = False

        messages.success(
            request=request,
            message=mark_safe(
                s=_(
                    f'<h4>Success!</h4><p>Topic "{curent_topic}" is no longer "sticky".</p>'  # pylint: disable=line-too-long
                )
            ),
        )

        logger.info(
            msg=f'{request.user} changed topic "{curent_topic}" to be no longer sticky.'
        )
    else:
        curent_topic.is_sticky = True

        messages.success(
            request=request,
            message=mark_safe(
                s=_(f'<h4>Success!</h4><p>Topic "{curent_topic}" is now "sticky".</p>')
            ),
        )

        logger.info(msg=f'{request.user} changed topic "{curent_topic}" to be sticky.')

    curent_topic.save(update_fields=["is_sticky"])

    return redirect(to=curent_topic.board.get_absolute_url())


@login_required
@permission_required(perm="aa_forum.manage_forum")
def topic_delete(request: WSGIRequest, topic_id: int) -> HttpResponseRedirect:
    """
    Delete a topic from a board

    :param request:
    :type request:
    :param topic_id:
    :type topic_id:
    :return:
    :rtype:
    """

    try:
        current_topic = Topic.objects.select_related("board").get(pk=topic_id)
    except Topic.DoesNotExist:
        return HttpResponseNotFound("Could not find topic.")

    topic__board = current_topic.board
    topic__subject = current_topic.subject

    current_topic.delete()

    messages.success(
        request=request,
        message=mark_safe(
            s=_(f'<h4>Success!</h4><p>Topic "{topic__subject}" removed.</p>')
        ),
    )

    logger.info(msg=f'{request.user} removed topic "{topic__subject}".')

    return redirect(to=topic__board.get_absolute_url())


@login_required
@permission_required(perm="aa_forum.basic_access")
def message(
    request: WSGIRequest,
    category_slug: str,  # pylint: disable=unused-argument
    board_slug: str,  # pylint: disable=unused-argument
    topic_slug: str,  # pylint: disable=unused-argument
    message_id: int,
) -> HttpResponseRedirect:
    """
    Redirect to a given message

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    try:
        current_message = Message.objects.select_related(
            "topic", "topic__board", "topic__board__category"
        ).get(pk=message_id)
    except Message.DoesNotExist:
        messages.error(
            request=request,
            message=mark_safe(s=_("<h4>Error!</h4><p>The message doesn't exist.</p>")),
        )

        return redirect(to="aa_forum:forum_index")

    return redirect(to=current_message.get_absolute_url())


@login_required
@permission_required(perm="aa_forum.basic_access")
def message_modify(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> HttpResponse:
    """
    Modify a message

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    message_to_modify = _message_from_slugs(
        request=request,
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        message_id=message_id,
    )

    # Check if the message exists.
    # If not, the user doesn't have access to its board, or it doesn't exist.
    # Either way, the message can't be edited
    if not message_to_modify:
        messages.error(
            request=request,
            message=mark_safe(
                s=_(
                    "<h4>Error!</h4><p>The message you were trying to modify does "
                    "either not exist, or you don't have access to it.</p>"
                )
            ),
        )

        logger.info(
            msg=(
                f"{request.user} trying to change a message in a topic that either "
                "does not exist or they have no access to."
            )
        )

        return redirect(to="aa_forum:forum_index")

    # Check if the user actually has the right to edit this message
    if (
        message_to_modify.user_created_id != request.user.id
        and not request.user.has_perm(perm="aa_forum.manage_forum")
    ):
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>You are not allowed to modify this message!</p>")
            ),
        )

        logger.info(
            msg=(
                f"{request.user} tried to modify a message in topic "
                f'"{message_to_modify.topic.subject}" without permission to do so.'
            )
        )

        return redirect(to=message_to_modify.topic.get_absolute_url())

    # We are in the clear, let's see what we've got
    if request.method == "POST":
        # Create a form instance and populate it with data from the request
        form = EditMessageForm(data=request.POST)

        # Check whether it's valid:
        if form.is_valid():
            message_to_modify.user_updated = request.user
            message_to_modify.message = form.cleaned_data["message"]
            message_to_modify.save()

            messages.success(
                request=request,
                message=mark_safe(
                    s=_("<h4>Success!</h4><p>The message has been updated.</p>")
                ),
            )

            logger.info(
                msg=(
                    f"{request.user} modified message ID {message_to_modify.pk} "
                    f'in topic "{message_to_modify.topic.subject}".'
                )
            )

            return redirect(
                to="aa_forum:forum_message",
                category_slug=category_slug,
                board_slug=board_slug,
                topic_slug=topic_slug,
                message_id=message_id,
            )

        # Form invalid
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>Mandatory form field is empty.</p>")
            ),
        )
    # If not, we'll fill the form with the information from the message object
    else:
        form = EditMessageForm(instance=message_to_modify)

    context = {
        "form": form,
        "board": message_to_modify.topic.board,
        "message": message_to_modify,
    }

    logger.info(
        msg=(
            f"{request.user} is modifying message ID {message_to_modify.pk} "
            f'in topic "{message_to_modify.topic.subject}".'
        )
    )

    return render(
        request=request,
        template_name="aa_forum/view/forum/modify-message.html",
        context=context,
    )


def _message_from_slugs(
    request: WSGIRequest,
    category_slug: str,
    board_slug: str,
    topic_slug: str,
    message_id: int,
) -> Message | None:
    """
    Helper function to get a message from slugs

    :param request:
    :type request:
    :param category_slug:
    :type category_slug:
    :param board_slug:
    :type board_slug:
    :param topic_slug:
    :type topic_slug:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    current_message = Message.objects.get_from_slugs(
        category_slug=category_slug,
        board_slug=board_slug,
        topic_slug=topic_slug,
        message_id=message_id,
        user=request.user,
    )

    return current_message


@login_required
@permission_required(perm="aa_forum.basic_access")
def message_delete(request: WSGIRequest, message_id: int) -> HttpResponseRedirect:
    """
    Delete a message from a topic

    :param request:
    :type request:
    :param message_id:
    :type message_id:
    :return:
    :rtype:
    """

    try:
        current_message = (
            Message.objects.select_related(
                "topic", "topic__board", "topic__board__category"
            )
            .user_has_access(user=request.user)
            .get(pk=message_id)
        )
    except Message.DoesNotExist:
        return HttpResponseNotFound("Message not found.")

    current_message__topic = current_message.topic
    current_message__topic__subject = current_message__topic.subject

    # Safety check to make sure the user is allowed to delete this message
    if (
        current_message.user_created_id != request.user.id
        and not request.user.has_perm(perm="aa_forum.manage_forum")
    ):
        messages.error(
            request=request,
            message=mark_safe(
                s=_("<h4>Error!</h4><p>You are not allowed to delete this message!</p>")
            ),
        )

        logger.info(
            msg=(
                f"{request.user} was trying to delete message ID {message_id} without "
                "permission to do so. Redirecting to forum index."
            )
        )

        return redirect(
            to="aa_forum:forum_topic",
            category_slug=current_message__topic.board.category.slug,
            board_slug=current_message__topic.board.slug,
            topic_slug=current_message__topic.slug,
        )

    # Let's check if we have more than one message in this topic
    # and the message we want to delete is not the first
    # If so, remove just that message and return to the topic
    if (
        current_message__topic.first_message.pk != current_message.pk
        and current_message__topic.messages.all().count() > 1
    ):
        current_message.delete()

        messages.success(
            request=request,
            message=mark_safe(
                s=_("<h4>Success!</h4><p>The message has been deleted.</p>")
            ),
        )

        logger.info(
            msg=(
                f"{request.user} removed message ID {message_id} "
                f'from topic "{current_message__topic__subject}".'
            )
        )

        return redirect(
            to="aa_forum:forum_topic",
            category_slug=current_message__topic.board.category.slug,
            board_slug=current_message__topic.board.slug,
            topic_slug=current_message__topic.slug,
        )

    # If it is the topics opening post …
    current_message__topic.delete()

    messages.success(
        request=request,
        message=mark_safe(
            s=_(
                "<h4>Success!</h4><p>The message has been deleted.</p>"
                "<p>This was the topics opening post, so the topic has been "
                "deleted as well.</p>"
            )
        ),
    )

    logger.info(
        msg=(
            f"{request.user} removed message ID {message_id}. This was the original "
            f'post, so the topic "{current_message__topic__subject}" has been '
            "removed as well."
        )
    )

    return redirect(
        to="aa_forum:forum_board",
        category_slug=current_message__topic.board.category.slug,
        board_slug=current_message__topic.board.slug,
    )


@login_required
@permission_required(perm="aa_forum.basic_access")
def mark_all_as_read(request: WSGIRequest) -> HttpResponseRedirect:
    """
    Mark all topics as read

    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )

    boards = (
        Board.objects.prefetch_related(
            Prefetch(
                lookup="topics",
                queryset=Topic.objects.select_related(
                    "last_message",
                ).annotate(has_unread_messages=~Exists(queryset=has_read_all_messages)),
            )
        )
        .user_has_access(user=request.user)
        .all()
    )

    if boards.count() > 0:
        for board_in_loop in boards:
            for topic_in_loop in board_in_loop.topics.all():
                LastMessageSeen.objects.update_or_create(
                    topic=topic_in_loop,
                    user=request.user,
                    defaults={"message_time": topic_in_loop.last_message.time_posted},
                )

    logger.info(msg=f"{request.user} marked all topics as read.")

    return redirect(to="aa_forum:forum_index")


@login_required
@permission_required(perm="aa_forum.basic_access")
def unread_topics_count(request: WSGIRequest) -> int:
    """
    Get the number of unread messages for the user

    :param request:
    :type request:
    :return:
    :rtype:
    """

    has_read_all_messages = LastMessageSeen.objects.filter(
        topic=OuterRef("pk"),
        user=request.user,
        message_time__gte=OuterRef("last_message__time_posted"),
    )
    unread_topic_pks = Topic.objects.filter(
        ~Exists(queryset=has_read_all_messages)
    ).values_list("pk", flat=True)

    boards = (
        Board.objects.annotate(
            num_unread_topics=Count(
                expression="topics",
                filter=Q(topics__in=unread_topic_pks),
                distinct=True,
            ),
        )
        .user_has_access(user=request.user)
        .all()
    )

    count_unread_topics = 0

    if boards.count() > 0:
        for board_in_loop in boards:
            count_unread_topics += board_in_loop.num_unread_topics

    return count_unread_topics



"""
Helper functions for forms
"""

# Django
from django.contrib import messages
from django.core.handlers.wsgi import WSGIRequest
from django.forms import Form
from django.utils.translation import gettext_lazy


def message_form_errors(request: WSGIRequest, form: Form) -> None:
    """
    Add form errors to messages

    :param request:
    :type request:
    :param form:
    :type form:
    :return:
    :rtype:
    """

    for _, errors in form.errors.items():
        for text in errors:
            messages.error(request=request, message=f"{gettext_lazy('Error:')} {text}")



"""
Constants
"""

# Standard Library
import glob
import json
import os

# Django
from django.utils.text import slugify

# AA Forum
from aa_forum import __version__

github_url: str = "https://github.com/ppfeufer/aa-forum"
verbose_name: str = "AA-Forum - A simple forum for Alliance Auth"
verbose_name_slug: str = slugify(value=verbose_name, allow_unicode=True)
user_agent: str = f"{verbose_name_slug} v{__version__} {github_url}"

# All internal URLs need to start with this prefix
# to prevent conflicts with user-generated forum URLs
INTERNAL_URL_PREFIX = "-"

# Default sort order for new categories and boards
DEFAULT_CATEGORY_AND_BOARD_SORT_ORDER = 999999

# Search stop words. These words and characters will be removed from the search phrase
SEARCH_STOPWORDS = ['"', "<", ">", "(", ")", "{", "}"]

# Get stopwords file list (Files downloaded from: https://github.com/stopwords-iso)
file_list = glob.glob(
    os.path.join(
        os.path.abspath(os.path.dirname(__file__)), "search/stopwords", "*.json"
    )
)

# Add stopwords from all files
for file in file_list:
    with open(file, encoding="utf-8") as f:
        SEARCH_STOPWORDS.extend(json.load(f))

# Discord embed settings
DISCORD_EMBED_COLOR_INFO = 0x5BC0DE
DISCORD_EMBED_COLOR_SUCCESS = 0x5CB85C
DISCORD_EMBED_COLOR_WARNING = 0xF0AD4E
DISCORD_EMBED_COLOR_DANGER = 0xD9534F

DISCORD_EMBED_COLOR_MAP = {
    "info": DISCORD_EMBED_COLOR_INFO,
    "success": DISCORD_EMBED_COLOR_SUCCESS,
    "warning": DISCORD_EMBED_COLOR_WARNING,
    "danger": DISCORD_EMBED_COLOR_DANGER,
}

DISCORD_EMBED_MESSAGE_LENGTH = 1000



"""
Tests for the forum views
"""

# Standard Library
from http import HTTPStatus
from unittest.mock import patch

# Django
from django.contrib.auth.models import Group
from django.contrib.messages import get_messages
from django.test import TestCase
from django.urls import reverse

# AA Forum
from aa_forum.models import Board, Category, LastMessageSeen, Message, Topic
from aa_forum.tests.utils import create_fake_messages, create_fake_user, my_get_setting

VIEWS_PATH = "aa_forum.views.forum"


class TestIndexViews(TestCase):
    """
    Test the forum views
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up users and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.user_1001 = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Peter Parker",
            permissions=["aa_forum.basic_access"],
        )
        cls.category = Category.objects.create(name="Science")

    def setUp(self) -> None:
        """
        Set up boards and topics

        :return:
        :rtype:
        """

        # board 1 has an unread topic
        self.board_1 = Board.objects.create(name="Physics", category=self.category)
        topic_1 = Topic.objects.create(subject="Mysteries", board=self.board_1)
        create_fake_messages(topic=topic_1, amount=4)
        topic_2 = Topic.objects.create(subject="Recent Discoveries", board=self.board_1)
        create_fake_messages(topic=topic_2, amount=2)
        LastMessageSeen.objects.create(
            topic=topic_2,
            user=self.user_1001,
            message_time=topic_2.messages.order_by("-time_posted")[0].time_posted,
        )

        # board 2 has no unread topics
        self.board_2 = Board.objects.create(name="Math", category=self.category)
        topic = Topic.objects.create(subject="Unsolved Problems", board=self.board_2)
        create_fake_messages(topic=topic, amount=2)
        LastMessageSeen.objects.create(
            topic=topic,
            user=self.user_1001,
            message_time=topic.messages.order_by("-time_posted")[0].time_posted,
        )

    def test_should_show_index(self):
        """
        Test should show forum index

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)

    def test_should_show_new_indicator_when_one_topic_not_seen_yet(self):
        """
        Test should show new indicator when a topic has not been seen yet

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-unread-in-{self.board_1.id}")
        self.assertNotContains(
            response=res, text=f"aa-forum-unread-in-{self.board_2.id}"
        )

    def test_should_show_new_indicator_when_new_posts_are_made(self):
        """
        Test should show new indicator when new posts are made

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        Message.objects.create(
            topic=self.board_2.topics.first(),
            user_created=self.user_1002,
            message="new message",
        )
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-unread-in-{self.board_2.id}")

    def test_should_show_counts(self):
        """
        Test should show counts

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text="6 Posts")
        self.assertContains(response=res, text="2 Topics")

    def test_should_show_empty_counts_after_all_topics_are_deleted(self):
        """
        Test should show empty counts after all topics are deleted

        :return:
        :rtype:
        """

        # given
        Topic.objects.all().delete()
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text="0 Posts")
        self.assertContains(response=res, text="0 Topics")


class TestIndexViewsSpecial(TestCase):
    """
    Test some special views
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up users and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.user_1001 = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Peter Parker",
            permissions=["aa_forum.basic_access"],
        )
        cls.category = Category.objects.create(name="Science")

    def test_should_show_empty_counts_after_all_topics_are_deleted_with_child_board(
        self,
    ):
        """
        Test should show empty counts after all topics are deleted with child board

        :return:
        :rtype:
        """

        # given
        board = Board.objects.create(name="Physics", category=self.category)
        topic = Topic.objects.create(subject="alpha", board=board)
        create_fake_messages(topic=topic, amount=1)
        child_board = Board.objects.create(
            name="Thermodynamics", category=self.category, parent_board=board
        )
        child_topic = Topic.objects.create(subject="bravo", board=child_board)
        create_fake_messages(topic=child_topic, amount=1)
        child_topic.delete()
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(path=reverse(viewname="aa_forum:forum_index"))

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text="1 Posts")
        self.assertContains(response=res, text="1 Topics")


class TestBoardViews(TestCase):
    """
    Test board views
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up users and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.user_1001 = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1002 = create_fake_user(
            character_id=1002,
            character_name="Peter Parker",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1003 = create_fake_user(
            character_id=1003,
            character_name="Clark Kent",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )
        cls.category = Category.objects.create(name="Science")
        cls.board = Board.objects.create(name="Physics", category=cls.category)

    def setUp(self) -> None:
        """
        Set up topics

        :return:
        :rtype:
        """

        # topic 1 is completely new
        self.topic_1 = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=self.topic_1, amount=15)
        # topic 2 is read
        self.topic_2 = Topic.objects.create(subject="Off Topic", board=self.board)
        create_fake_messages(topic=self.topic_2, amount=9)
        LastMessageSeen.objects.create(
            topic=self.topic_2,
            user=self.user_1001,
            message_time=self.topic_2.messages.order_by("-time_posted")[0].time_posted,
        )

    def test_should_show_new_indicator_when_topic_not_seen_yet(self):
        """
        Test should show new indicator when topic has not been seen yet

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.board.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-link-new-{self.topic_1.id}")

    def test_should_show_new_indicator_when_seen_first_page_only(self):
        """
        Test should show new indicator when first page seen only

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        last_message_seen = self.topic_1.messages.order_by("time_posted")[4]
        LastMessageSeen.objects.create(
            topic=self.topic_1,
            user=self.user_1001,
            message_time=last_message_seen.time_posted,
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.board.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-link-new-{self.topic_1.id}")

    def test_should_show_new_indicator_when_seen_second_page_only(self):
        """
        Test should show new indicator when only second page been seen

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        last_message_seen = self.topic_1.messages.order_by("time_posted")[9]
        LastMessageSeen.objects.create(
            topic=self.topic_1,
            user=self.user_1001,
            message_time=last_message_seen.time_posted,
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.board.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-link-new-{self.topic_1.id}")

    def test_should_not_show_new_indicator_when_seen_last_page(self):
        """
        Test should show new indicator when only last page been seen

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        last_message_seen = self.topic_1.messages.order_by("time_posted")[14]
        LastMessageSeen.objects.create(
            topic=self.topic_1,
            user=self.user_1001,
            message_time=last_message_seen.time_posted,
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.board.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertNotContains(
            response=res, text=f"aa-forum-link-new-{self.topic_1.id}"
        )

    def test_should_show_new_indicator_when_new_posts_are_made(self):
        """
        Test should show new indicator when new posts are made

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        Message.objects.create(
            topic=self.topic_2, user_created=self.user_1002, message="new message"
        )
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.board.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        self.assertContains(response=res, text=f"aa-forum-link-new-{self.topic_2.id}")

    def test_should_delete_topic(self):
        """
        Test should delete a topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_topic_delete", args=[self.topic_1.pk])
        )

        # then
        self.assertRedirects(
            response=res,
            expected_url=reverse(
                viewname="aa_forum:forum_board",
                args=[self.board.category.slug, self.board.slug],
            ),
        )
        self.assertFalse(expr=self.board.topics.filter(pk=self.topic_1.pk).exists())

    def test_should_return_404_when_delete_topic_not_found(self):
        """
        Test should return 404 when a topic not found on delete

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_topic_delete", args=[0])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.NOT_FOUND)

    def test_should_lock_topic(self):
        """
        Test should lock a topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_change_lock_state",
                args=[self.topic_1.pk],
            )
        )

        # then
        self.assertRedirects(
            response=res,
            expected_url=reverse(
                viewname="aa_forum:forum_board",
                args=[self.board.category.slug, self.board.slug],
            ),
        )
        self.topic_1.refresh_from_db()
        self.assertTrue(expr=self.topic_1.is_locked)

    def test_should_unlock_topic(self):
        """
        Test should unlock a topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)
        self.topic_1.is_locked = True
        self.topic_1.save()

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_change_lock_state",
                args=[self.topic_1.pk],
            )
        )

        # then
        self.assertRedirects(
            response=res,
            expected_url=reverse(
                viewname="aa_forum:forum_board",
                args=[self.board.category.slug, self.board.slug],
            ),
        )
        self.topic_1.refresh_from_db()
        self.assertFalse(expr=self.topic_1.is_locked)

    def test_should_return_404_when_lock_topic_not_found(self):
        """
        Test should return 404 when a topic not found on lock

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_topic_change_lock_state", args=[0])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.NOT_FOUND)

    def test_should_make_topic_sticky(self):
        """
        Test should make topic sticky

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_change_sticky_state",
                args=[self.topic_1.pk],
            )
        )

        # then
        self.assertRedirects(
            response=res,
            expected_url=reverse(
                viewname="aa_forum:forum_board",
                args=[self.board.category.slug, self.board.slug],
            ),
        )
        self.topic_1.refresh_from_db()
        self.assertTrue(expr=self.topic_1.is_sticky)

    def test_should_reverse_topic_sticky(self):
        """
        Test should reverse topic sticky

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)
        self.topic_1.is_sticky = True
        self.topic_1.save()

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_change_sticky_state",
                args=[self.topic_1.pk],
            )
        )

        # then
        self.assertRedirects(
            response=res,
            expected_url=reverse(
                viewname="aa_forum:forum_board",
                args=[self.board.category.slug, self.board.slug],
            ),
        )
        self.topic_1.refresh_from_db()
        self.assertFalse(expr=self.topic_1.is_sticky)

    def test_should_return_404_when_sticky_topic_not_found(self):
        """
        Test should return 404 when a topic not found on sticky change

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_topic_change_sticky_state", args=[0])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.NOT_FOUND)

    def test_should_return_board_does_not_exist_for_wrong_board_on_board_view(self):
        """
        Test should return "Board does not exist" for wrong board

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board",
                args=["foo", "bar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertRaises(expected_exception=Board.DoesNotExist)
        self.assertEqual(
            first=response.url, second=reverse(viewname="aa_forum:forum_index")
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The board you were trying to visit does "
                "either not exist, or you don't have access to it.</p>"
            ),
        )

    def test_should_return_category_does_not_exists_on_new_topic(self):
        """
        Test should return "Category does not exist" on new topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=["foo", "bar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertRaises(expected_exception=Category.DoesNotExist)
        self.assertEqual(
            first=response.url, second=reverse(viewname="aa_forum:forum_index")
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The category you were trying to post in does "
                "not exist.</p>"
            ),
        )

    def test_should_return_board_does_not_exists_on_new_topic(self):
        """
        Test should return "Board does not exist" on new topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=[self.category.slug, "bar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertRaises(expected_exception=Board.DoesNotExist)
        self.assertEqual(
            first=response.url, second=reverse(viewname="aa_forum:forum_index")
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The board you were trying to post in does "
                "either not exist, or you don't have access to it.</p>"
            ),
        )


@patch(VIEWS_PATH + ".Setting.objects.get_setting", new=my_get_setting)
class TestTopicViews(TestCase):
    """
    Test topic views
    """

    @classmethod
    def setUpClass(cls) -> None:
        """
        Set up users and categories

        :return:
        :rtype:
        """

        super().setUpClass()
        cls.user_1001 = create_fake_user(
            character_id=1001,
            character_name="Bruce Wayne",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1003 = create_fake_user(
            character_id=1003,
            character_name="Clark Lent",
            permissions=["aa_forum.basic_access", "aa_forum.manage_forum"],
        )
        cls.group = Group.objects.create(name="Superhero")
        cls.announcement_group = Group.objects.create(name="Justice League")

        cls.user_1004 = create_fake_user(
            character_id=1004,
            character_name="Luke Skywalker",
            permissions=["aa_forum.basic_access"],
        )
        cls.user_1004.groups.add(cls.announcement_group)

    def setUp(self) -> None:
        """
        Set up categories, boards and topics

        :return:
        :rtype:
        """

        self.category = Category.objects.create(name="Science")
        self.board = Board.objects.create(name="Physics", category=self.category)
        self.announcement_board = Board.objects.create(
            name="Chemistry", category=self.category, is_announcement_board=True
        )
        self.topic = Topic.objects.create(subject="Mysteries", board=self.board)
        create_fake_messages(topic=self.topic, amount=15)

    def test_should_remember_last_message_seen_by_user_page_1(self):
        """
        Test should remember the last message seen by user on page 1

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        last_message_seen = LastMessageSeen.objects.get(
            topic=self.topic, user=self.user_1001
        )

        # View has 2 pages á 5 messages. This is last message on 1st page
        last_message = self.topic.messages.order_by("time_posted")[4]
        self.assertEqual(
            first=last_message_seen.message_time, second=last_message.time_posted
        )

    def test_should_remember_last_message_seen_by_user_page_2(self):
        """
        Test should remember the last message seen by user on page 2

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic",
                args=[self.category.slug, self.board.slug, self.topic.slug, 2],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        last_message_seen = LastMessageSeen.objects.get(
            topic=self.topic, user=self.user_1001
        )

        # View has 2 pages á 5 messages. This is last message on 2nd page
        last_message = Message.objects.order_by("time_posted")[9]
        self.assertEqual(
            first=last_message_seen.message_time, second=last_message.time_posted
        )

    def test_should_remember_last_message_seen_by_user_when_opening_previous_pages(
        self,
    ):
        """
        Test should remember the last message seen by user when opening previous pages

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic",
                args=[self.category.slug, self.board.slug, self.topic.slug, 2],
            )
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic",
                args=[self.category.slug, self.board.slug, self.topic.slug, 1],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.OK)
        last_message_seen = LastMessageSeen.objects.get(
            topic=self.topic, user=self.user_1001
        )

        # View has 2 pages á 5 messages. This is last message on 2nd page
        last_message = Message.objects.order_by("time_posted")[9]
        self.assertEqual(
            first=last_message_seen.message_time, second=last_message.time_posted
        )

    def test_should_redirect_to_first_message_when_topic_not_seen_yet(self):
        """
        Test should redirect to the first message when the topic not seen yet

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        first_message = self.topic.messages.order_by("time_posted").first()

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_first_unread_message",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(
            response=res, expected_url=first_message.get_absolute_url()
        )

    def test_should_redirect_to_first_new_message_normal(self):
        """
        Test should normally redirect to the first new message

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        messages_sorted = list(self.topic.messages.order_by("time_posted"))
        last_seen_message = messages_sorted[2]
        first_unseen_message = messages_sorted[3]
        LastMessageSeen.objects.create(
            topic=self.topic,
            user=self.user_1001,
            message_time=last_seen_message.time_posted,
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_first_unread_message",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(
            response=res, expected_url=first_unseen_message.get_absolute_url()
        )

    def test_should_redirect_to_first_unseen_message_when_last_seen_message_deleted(
        self,
    ):
        """
        Test should redirect to the first unseen message when the last seen message is deleted

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        messages_sorted = list(self.topic.messages.order_by("time_posted"))
        last_seen_message = messages_sorted[2]
        first_unseen_message = messages_sorted[3]
        LastMessageSeen.objects.create(
            topic=self.topic,
            user=self.user_1001,
            message_time=last_seen_message.time_posted,
        )
        last_seen_message.delete()

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_first_unread_message",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(
            response=res, expected_url=first_unseen_message.get_absolute_url()
        )

    def test_should_redirect_to_newest_message_when_seen_full_topic(self):
        """
        Test should redirect to the newest message

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        last_seen_message = self.topic.messages.order_by("-time_posted")[0]
        LastMessageSeen.objects.create(
            topic=self.topic,
            user=self.user_1001,
            message_time=last_seen_message.time_posted,
        )

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_first_unread_message",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            )
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(
            response=res, expected_url=last_seen_message.get_absolute_url()
        )

    def test_should_redirect_to_message_by_id_first_page(self):
        """
        Test should redirect to message by ID on the first page

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        my_message = self.topic.messages.order_by("time_posted")[3]
        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_message",
                args=[
                    my_message.topic.board.category.slug,
                    my_message.topic.board.slug,
                    my_message.topic.slug,
                    my_message.pk,
                ],
            ),
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(response=res, expected_url=my_message.get_absolute_url())

    def test_should_redirect_to_message_by_id_middle_page_1(self):
        """
        Test should redirect to message by ID on page 1

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        my_message = self.topic.messages.order_by("time_posted")[5]

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_message",
                args=[
                    my_message.topic.board.category.slug,
                    my_message.topic.board.slug,
                    my_message.topic.slug,
                    my_message.pk,
                ],
            ),
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(response=res, expected_url=my_message.get_absolute_url())

    def test_should_redirect_to_message_by_id_middle_page_2(self):
        """
        Test should redirect to message by ID on page 2

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        my_message = self.topic.messages.order_by("time_posted")[9]

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_message",
                args=[
                    my_message.topic.board.category.slug,
                    my_message.topic.board.slug,
                    my_message.topic.slug,
                    my_message.pk,
                ],
            ),
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(response=res, expected_url=my_message.get_absolute_url())

    def test_should_redirect_to_message_by_id_last_page(self):
        """
        Test should redirect to message by ID on the last page

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        my_message = self.topic.messages.order_by("time_posted")[12]

        # when
        res = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_message",
                args=[
                    my_message.topic.board.category.slug,
                    my_message.topic.board.slug,
                    my_message.topic.slug,
                    my_message.pk,
                ],
            ),
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertRedirects(response=res, expected_url=my_message.get_absolute_url())

    def test_should_delete_regular_message(self):
        """
        Test should delete a regular message

        :return:
        :rtype:

        """

        # given
        self.client.force_login(user=self.user_1003)
        my_message = self.topic.messages.last()

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_message_delete", args=[my_message.pk])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second=self.topic.get_absolute_url())
        self.assertFalse(expr=self.topic.messages.filter(pk=my_message.pk).exists())

    def test_should_not_delete_message_because_missing_permissions(self):
        """
        Test should not delete a message because of insufficient permissions

        :return:
        :rtype:

        """

        # given
        message = Message(
            message="Lorem Ipsum", topic=self.topic, user_created=self.user_1003
        )
        message.save()

        # when
        self.client.force_login(user=self.user_1001)
        response = self.client.get(
            path=reverse(viewname="aa_forum:forum_message_delete", args=[message.pk])
        )

        # then
        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)

        expected_message = (
            "<h4>Error!</h4><p>You are not allowed to delete this message!</p>"
        )
        messages = list(get_messages(response.wsgi_request))

        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(first=str(messages[0]), second=expected_message)

    def test_should_delete_first_message(self):
        """
        Test should delete the first message

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)
        my_message = self.topic.messages.first()

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_message_delete", args=[my_message.pk])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second=self.topic.board.get_absolute_url())
        self.assertFalse(expr=self.topic.messages.filter(pk=my_message.pk).exists())
        self.assertFalse(expr=self.board.topics.filter(pk=self.topic.pk).exists())

    def test_should_delete_last_message_in_topic(self):
        """
        Test should delete the last message in a topic

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)
        my_message = self.topic.messages.first()
        self.topic.messages.exclude(pk=my_message.pk).delete()

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_message_delete", args=[my_message.pk])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second=self.topic.board.get_absolute_url())
        self.assertFalse(expr=self.topic.messages.filter(pk=my_message.pk).exists())
        self.assertFalse(expr=self.board.topics.filter(pk=self.topic.pk).exists())

    def test_should_return_404_when_delete_message_not_found(self):
        """
        Test should return 404 when a message not found on delete

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        res = self.client.get(
            path=reverse(viewname="aa_forum:forum_message_delete", args=[0])
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.NOT_FOUND)

    @patch(VIEWS_PATH + ".messages")
    def test_should_not_edit_message_from_others(self, messages):
        """
        Test should not be able to edit messages from others

        :param messages:
        :type messages:
        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)
        alien_message = Message.objects.create(
            topic=self.topic, user_created=self.user_1003, message="old text"
        )

        # when
        res = self.client.post(
            path=reverse(
                viewname="aa_forum:forum_message_modify",
                args=[
                    self.category.slug,
                    self.board.slug,
                    self.topic.slug,
                    alien_message.pk,
                ],
            ),
            data={"message": "new text"},
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second=self.topic.get_absolute_url())

        alien_message.refresh_from_db()
        self.assertEqual(first=alien_message.message, second="old text")
        self.assertTrue(expr=messages.error.called)

    @patch(VIEWS_PATH + ".messages")
    def test_should_not_edit_message_from_board_with_no_access(self, messages):
        """
        Test should not be able to edit messages from boards with no access

        :param messages:
        :type messages:
        :return:
        :rtype:
        """

        # given
        self.board.groups.add(self.group)
        self.client.force_login(user=self.user_1001)
        alien_message = Message.objects.create(
            topic=self.topic, user_created=self.user_1001, message="old text"
        )

        # when
        res = self.client.post(
            path=reverse(
                viewname="aa_forum:forum_message_modify",
                args=[
                    self.category.slug,
                    self.board.slug,
                    self.topic.slug,
                    alien_message.pk,
                ],
            ),
            data={"message": "new text"},
        )

        # then
        self.assertEqual(first=res.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=res.url, second=reverse("aa_forum:forum_index"))

        alien_message.refresh_from_db()
        self.assertEqual(first=alien_message.message, second="old text")
        self.assertTrue(expr=messages.error.called)

    def test_should_return_redirect_to_forum_index_if_topic_does_not_exist_on_topic_view(  # pylint: disable=line-too-long
        self,
    ):
        """
        Test should redirect to forum index when the topic does not exist on topic view

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic",
                args=["foo", "bar", "foobar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=response.url, second=reverse("aa_forum:forum_index"))
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The topic you were trying to view does not "
                "exist or you do not have access to it.</p>"
            ),
        )

    def test_should_return_redirect_to_forum_index_if_topic_does_not_exist_on_topic_modify_view(  # pylint: disable=line-too-long
        self,
    ):
        """
        Test should redirect to forum index when topic does not exist on topic modify

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_modify",
                args=["foo", "bar", "foobar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=response.url, second=reverse("aa_forum:forum_index"))
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The topic you were trying to modify does not "
                "exist or you do not have access to it.</p>"
            ),
        )

    def test_should_return_redirect_to_forum_index_if_topic_does_not_exist_on_topic_reply(  # pylint: disable=line-too-long
        self,
    ):
        """
        Test should redirect to forum index if a topic does not exist on reply

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_reply",
                args=["foo", "bar", "foobar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=response.url, second=reverse("aa_forum:forum_index"))
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The topic you were trying to reply does not "
                "exist or you do not have access to it.</p>"
            ),
        )

    def test_should_return_redirect_to_forum_index_if_message_does_not_exist_on_message_view(  # pylint: disable=line-too-long
        self,
    ):
        """
        Test should redirect to forum index if a message does not exist

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_message",
                args=["foo", "bar", "foobar", 0],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=response.url, second=reverse("aa_forum:forum_index"))
        self.assertRaises(expected_exception=Message.DoesNotExist)
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second="<h4>Error!</h4><p>The message doesn't exist.</p>",
        )

    def test_should_show_all_unread_messages_view(self):
        """
        Test should show all unread messages

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(viewname="aa_forum:forum_topic_show_all_unread"),
        )

        # then
        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    def test_should_return_to_forum_index_on_topic_modify_when_no_topic_found(self):
        """
        Test should redirect to forum index if a topic not found on modify

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_modify",
                args=["foo", "bar", "foobar"],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(first=response.url, second=reverse("aa_forum:forum_index"))
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The topic you were trying to modify does not "
                "exist or you do not have access to it.</p>"
            ),
        )

    def test_should_redirect_to_topic_view_for_user_without_rights_to_modify_topic(
        self,
    ):
        """
        Test should redirect to forum index for user without rights to modify topic

        :return:
        :rtype:
        """

        # given
        user_without_modify_perms = create_fake_user(
            character_id=1002,
            character_name="Peter Parker",
            permissions=["aa_forum.basic_access"],
        )
        self.client.force_login(user=user_without_modify_perms)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_modify",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            ),
        )

        # then
        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=response.url,
            second=reverse(
                "aa_forum:forum_topic",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            ),
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second="<h4>Error!</h4><p>You are not allowed to modify this topic!</p>",
        )

    def test_should_show_modify_topic_view(self):
        """
        Test should show modify topic view

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_topic_modify",
                args=[self.category.slug, self.board.slug, self.topic.slug],
            ),
        )

        # then
        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    def test_can_create_new_topic_in_announcement_board_with_permission(self):
        """
        Test should create a new topic in announcement board with permissions

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1003)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )

        # then
        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    def test_can_create_new_topic_in_announcement_board_with_group(self):
        """
        Test should create a new topic in announcement board with groups

        :return:
        :rtype:
        """

        # given
        self.announcement_board.announcement_groups.add(self.announcement_group)
        self.client.force_login(user=self.user_1004)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )

        # then
        self.assertEqual(first=response.status_code, second=HTTPStatus.OK)

    def test_cannot_create_new_topic_in_announcement_board_without_permission(self):
        """
        Test should create a new topic in announcement board without permissions

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1004)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )

        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=response.url,
            second=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The board you were trying to post in is an "
                "announcement board and you don't have the permissions to start a "
                "topic there.</p>"
            ),
        )

    def test_cannot_create_new_topic_in_announcement_board(self):
        """
        Test cannot create a new topic in announcement board

        :return:
        :rtype:
        """

        # given
        self.client.force_login(user=self.user_1001)

        # when
        response = self.client.get(
            path=reverse(
                viewname="aa_forum:forum_board_new_topic",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )

        messages = list(get_messages(request=response.wsgi_request))

        self.assertEqual(first=response.status_code, second=HTTPStatus.FOUND)
        self.assertEqual(
            first=response.url,
            second=reverse(
                viewname="aa_forum:forum_board",
                args=[self.category.slug, self.announcement_board.slug],
            ),
        )
        self.assertEqual(first=len(messages), second=1)
        self.assertEqual(
            first=str(messages[0]),
            second=(
                "<h4>Error!</h4><p>The board you were trying to post in is an "
                "announcement board and you don't have the permissions to start a "
                "topic there.</p>"
            ),
        )



"""
Hooks for the auth app
"""

# Alliance Auth
from allianceauth import hooks
from allianceauth.hooks import DashboardItemHook
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA Forum
from aa_forum import __title__, urls
from aa_forum.views.forum import unread_topics_count
from aa_forum.views.widgets import dashboard_widgets


class AaForumMenuItem(MenuItemHook):  # pylint: disable=too-few-public-methods
    """
    Menu item hook for AA Forum
    """

    def __init__(self):
        """
        Constructor
        """

        self.count = None

        # Setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            text=__title__,
            classes="fa-solid fa-comments",
            url_name="aa_forum:forum_index",
            navactive=["aa_forum:"],
        )

    def get_unread_topics_count(self, request):
        """
        Get the unread topics count

        :param request:
        :type request:
        :return:
        :rtype:
        """

        count_unread_topics = unread_topics_count(request=request)
        self.count = (
            count_unread_topics
            if count_unread_topics and count_unread_topics > 0
            else None
        )

    def render(self, request):
        """
        Render the menu item

        :param request:
        :type request:
        :return:
        :rtype:
        """

        if request.user.has_perm(perm="aa_forum.basic_access"):
            # Get the unread topics count
            self.get_unread_topics_count(request=request)

            # Render the menu item
            return MenuItemHook.render(self, request=request)

        return ""


class AaForumDashboardWidgets(DashboardItemHook):
    """
    This class adds the AA Forum dashboard widgets to the Alliance Auth dashboard.
    """

    def __init__(self):
        # Setup dashboard widget
        DashboardItemHook.__init__(self=self, view_function=dashboard_widgets, order=5)


@hooks.register("menu_item_hook")
def register_menu():
    """
    Register our menu item

    :return:
    :rtype:
    """

    return AaForumMenuItem()


@hooks.register("url_hook")
def register_urls():
    """
    Register our urls

    :return:
    :rtype:
    """

    return UrlHook(urls=urls, namespace="aa_forum", base_url=r"^forum/")


@hooks.register("dashboard_hook")
def register_dashboard_widgets():
    """
    Register our dashboard hook

    :return: The hook
    :rtype: AaForumDashboardWidgets
    """

    return AaForumDashboardWidgets()



"""
Pagination helper functions
"""

# Django
from django.core.paginator import Page, Paginator
from django.db.models.query import QuerySet


def get_paginated_page_object(
    queryset: QuerySet, items_per_page: int = 10, page_number: int = None
) -> Page:
    """
    Return a paginated page object

    :param queryset:
    :type queryset:
    :param items_per_page:
    :type items_per_page:
    :param page_number:
    :type page_number:
    :return:
    :rtype:
    """

    paginator = Paginator(object_list=queryset, per_page=items_per_page)
    page_obj = paginator.get_page(number=page_number)

    return page_obj



# Generated by Django 4.2.11 on 2024-03-08 15:28

# Django
from django.db import migrations, models

# CKEditor
import django_ckeditor_5.fields


class Migration(migrations.Migration):

    dependencies = [
        ("aa_forum", "0016_fix_quotation_marks"),
    ]

    operations = [
        migrations.AddField(
            model_name="userprofile",
            name="show_unread_topics_dashboard_widget",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="message",
            name="message",
            field=django_ckeditor_5.fields.CKEditor5Field(),
        ),
        migrations.AlterField(
            model_name="personalmessage",
            name="message",
            field=django_ckeditor_5.fields.CKEditor5Field(),
        ),
        migrations.AlterField(
            model_name="userprofile",
            name="signature",
            field=django_ckeditor_5.fields.CKEditor5Field(blank=True),
        ),
    ]
